/*
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * 
 * Contributor(s): Joseph Gilbert
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

#include "euler.h"
#include "Types.h"
#include "BLI_winstuff.h"

/*****************************/
//    Euler Python Object   
/*****************************/

struct PyMethodDef Euler_methods[] = {
	{"zero",(PyCFunction)Euler_Zero, METH_NOARGS,
				Euler_Zero_doc},
	{"canonical",(PyCFunction)Euler_Canonical, METH_NOARGS,
				Euler_Canonical_doc},
	{"toMatrix",(PyCFunction)Euler_ToMatrix, METH_NOARGS,
				Euler_ToMatrix_doc},
	{"toQuat",(PyCFunction)Euler_ToQuat, METH_NOARGS,
				Euler_ToQuat_doc},
	{NULL, NULL, 0, NULL}
};

//euler methods
static PyObject *Euler_ToQuat(EulerObject *self, PyObject *args)
{
	float *quat;
	
	quat = PyMem_Malloc(4*sizeof(float));
	QuatToEul(quat, self->vec);
	
	return newQuaternionObject(quat);
}


static PyObject *Euler_ToMatrix(EulerObject *self, PyObject *args)
{
	float *mat;

	mat = PyMem_Malloc(3*3*sizeof(float));

	//assumes object->worldspace rotation in euler
	EulToMat3(self->vec, mat);
	
	return newMatrixObject(mat,3,3);
}

static PyObject *Euler_Canonical(EulerObject *self, PyObject *args)
{
	float angle0, angle1, angle2;

	angle0 = self->vec[0];
	angle1 = self->vec[1];
	angle2 = self->vec[2];
	
	//convert to radians
	angle0 = angle0 * (M_PI/180);
	angle1 = angle1 * (M_PI/180);
	angle2 = angle2 * (M_PI/180);

	//canonical y
	angle1 += M_PI;
	angle1 -= (floor(((double)(angle1)) * (1.0f/(M_PI*2.0f)))) *
		(M_PI * 2.0f);
	angle1 -= M_PI;

	//test y axis
	if(angle1 < -(1.0f/(M_PI*2.0f))){
		angle1 = -M_PI - angle1;
		angle0 += M_PI;
		angle2 += M_PI;
	}else if (angle1 > (1.0f/(M_PI*2.0f))){
		angle1 = M_PI - angle1;
		angle0 += M_PI;
		angle2 += M_PI;
	}

	if(fabs((double)(angle1)) > ((1.0f/(M_PI*2.0f)) - 1e-4)){
		//gimbal lock
		angle0 += angle2;
		angle2 = 0.0f;
	}else{
		//not gimbal lcok
		//canonical z
		angle2 += M_PI;
		angle2 -= (floor(((double)(angle2)) * (1.0f/(M_PI*2.0f)))) *
			(M_PI * 2.0f);
		angle2 -= M_PI;
	}

	//canonical x
	angle0 += M_PI;
	angle0 -= (floor(((double)(angle0)) * (1.0f/(M_PI*2.0f)))) *
		(M_PI * 2.0f);
	angle0 -= M_PI;
	
	//back to degrees
	self->vec[0] = angle0 * (180/M_PI);
	self->vec[1] = angle1 * (180/M_PI);
	self->vec[2] = angle2 * (180/M_PI);

	return EXPP_incr_ret(Py_None);
}

static PyObject *Euler_Zero(EulerObject *self, PyObject *args)
{
	self->vec[0] = 0.0;
	self->vec[1] = 0.0;
	self->vec[2] = 0.0;

	return EXPP_incr_ret(Py_None);
}

static void Euler_dealloc(EulerObject *self)
{
  PyObject_DEL (self);
}

static PyObject *Euler_getattr(EulerObject *self, char *name)
{
	if (ELEM3(name[0], 'x', 'y', 'z') && name[1]==0){
		return PyFloat_FromDouble(self->vec[name[0]-'x']);
	}
 
	return Py_FindMethod(Euler_methods, (PyObject*)self, name);
}

static int Euler_setattr(EulerObject *self, char *name, PyObject *v)
{
	float val;

	if (!PyArg_Parse(v, "f", &val))
		return EXPP_ReturnIntError(PyExc_TypeError,	
		"unable to parse float argument\n");

	if (ELEM3(name[0], 'x', 'y', 'z') && name[1]==0)
		self->vec[name[0]-'x']= val;
	else return -1;
}

/* Eulers Sequence methods */
static PyObject *Euler_item(EulerObject *self, int i)
{
	if (i < 0 || i >= 3)
	  return EXPP_ReturnPyObjError (PyExc_IndexError, "array index out of range\n");

	return Py_BuildValue("f", self->vec[i]);

}

static PyObject *Euler_slice(EulerObject *self, int begin, int end)
{
	PyObject *list;
	int count;
  
	if (begin < 0) begin= 0;
	if (end > 3) end= 3;
	if (begin > end) begin= end;

	list= PyList_New(end-begin);

	for (count = begin; count < end; count++){
		PyList_SetItem(list, count-begin, PyFloat_FromDouble(self->vec[count]));
	}

	return list;
}

static int Euler_ass_item(EulerObject *self, int i, PyObject *ob)
{
	int temp;

	if (i < 0 || i >= 3)
		return EXPP_ReturnIntError(PyExc_IndexError,
					"array assignment index out of range\n");

	if (!PyNumber_Check(ob))
		return EXPP_ReturnIntError(PyExc_IndexError,
					"Euler member must be a number\n");

	if(!PyFloat_Check(ob)){
		if(!PyInt_Check(ob)){
			return EXPP_ReturnIntError(PyExc_TypeError,"int or float expected\n");
		}else{
			temp = PyInt_AsLong(ob);
			self->vec[i] = ((float)temp);
		}
	}else{
		self->vec[i]= PyFloat_AsDouble(ob);
	}

	return 0;
}

static int Euler_ass_slice(EulerObject *self, int begin, int end, PyObject *seq)
{
	int count;
	int tempInt;

	if (begin < 0) begin= 0;
	if (end > 3) end= 3;
	if (begin > end) begin= end;

	if (!PySequence_Check(seq))
		return EXPP_ReturnIntError(PyExc_TypeError,
					"illegal argument type for built-in operation\n");

	if (PySequence_Length(seq) != (end - begin))
		return EXPP_ReturnIntError(PyExc_TypeError,
					"size mismatch in slice assignment\n");

	for (count = begin; count < end; count++) {
		PyObject *ob = PySequence_GetItem(seq, count);

		if(!EXPP_check_sequence_consistency(ob, &PyFloat_Type)){
			if(!EXPP_check_sequence_consistency(ob, &PyInt_Type)){
				return -1;
			}else{
				if (!PyArg_Parse(ob, "i", &tempInt)) {
					self->vec[count] = ((float)tempInt);
					Py_DECREF(ob);
					return -1;
				}
			}
		}else{
			if (!PyArg_Parse(ob, "f", &self->vec[count])) {
				Py_DECREF(ob);
				return -1;
			}
		}

		Py_DECREF(ob);
	}

  return 0;
}

static PyObject *Euler_repr (EulerObject *self)
{
	int i, maxindex = 3 - 1;
	char ftoa[24];
	PyObject *str1, *str2;
	int x;

	str1 = PyString_FromString ("[");

	for (i = 0; i < maxindex; i++) {
		sprintf(ftoa, "%.3f, ", self->vec[i]);
		str2 = PyString_FromString (ftoa);
		if (!str1 || !str2) goto error;
		PyString_ConcatAndDel (&str1, str2);
	}

	sprintf(ftoa, "%.3f]\n", self->vec[maxindex]);
	str2 = PyString_FromString (ftoa);
	if (!str1 || !str2) goto error; 
	PyString_ConcatAndDel (&str1, str2);

	if (str1) return str1;

error:
	Py_XDECREF (str1);
	Py_XDECREF (str2);
	return EXPP_ReturnPyObjError (PyExc_MemoryError,
			"couldn't create PyString!\n");
}

static PySequenceMethods Euler_SeqMethods =
{
	(inquiry)       0,                     /* sq_length */
	(binaryfunc)    0,                     /* sq_concat */
	(intargfunc)    0,                     /* sq_repeat */
	(intargfunc)    Euler_item,            /* sq_item */
	(intintargfunc)   Euler_slice,         /* sq_slice */
	(intobjargproc)   Euler_ass_item,      /* sq_ass_item */
	(intintobjargproc)  Euler_ass_slice,   /* sq_ass_slice */
};

PyTypeObject euler_Type =
{
  PyObject_HEAD_INIT(NULL)
  0,                           /*ob_size*/
  "Euler",                     /*tp_name*/
  sizeof(EulerObject),         /*tp_basicsize*/
  0,                           /*tp_itemsize*/
  /* methods */
  (destructor)  Euler_dealloc,  /*tp_dealloc*/
  (printfunc)   0,              /*tp_print*/
  (getattrfunc) Euler_getattr, /*tp_getattr*/
  (setattrfunc) Euler_setattr, /*tp_setattr*/
  0,                           /*tp_compare*/
  (reprfunc)    Euler_repr,    /*tp_repr*/
  0,                           /*tp_as_number*/
  &Euler_SeqMethods,           /*tp_as_sequence*/
};

PyObject *newEulerObject(float *vec)
{
  EulerObject *self;
  int x;

  euler_Type.ob_type = &PyType_Type;

  self = PyObject_NEW(EulerObject, &euler_Type);

  if(!vec){
	  self->vec = PyMem_Malloc (3*sizeof (float));
	  for(x = 0; x < 3; x++){
		  self->vec[x] = 0.0f;
	  }
  }else	self->vec = vec;
 
  return (PyObject*) self;
}

