/*
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * Contributor(s): Michel Selten & Joseph Gilbert
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

 
#include "vector.h"
#include "matrix.h"
#include "BLI_arithb.h"

struct PyMethodDef Matrix_methods[] = {
	{"zero",(PyCFunction)Matrix_Zero, METH_NOARGS,
				Matrix_Zero_doc},
	{"identity",(PyCFunction)Matrix_Identity, METH_NOARGS,
				Matrix_Identity_doc},
	{"transpose",(PyCFunction)Matrix_Transpose, METH_NOARGS,
				Matrix_Transpose_doc},
	{"determinant",(PyCFunction)Matrix_Determinant, METH_NOARGS,
				Matrix_Determinant_doc},
	{"inverse",(PyCFunction)Matrix_Inverse, METH_NOARGS,
				Matrix_Inverse_doc},
	{"translationPart",(PyCFunction)Matrix_TranslationPart, METH_NOARGS,
				Matrix_TranslationPart_doc},
	{"rotationPart",(PyCFunction)Matrix_RotationPart, METH_NOARGS,
				Matrix_RotationPart_doc},
	{"resize4x4",(PyCFunction)Matrix_Resize4x4, METH_NOARGS,
				Matrix_Resize4x4_doc},
	{"toEuler",(PyCFunction)Matrix_toEuler, METH_NOARGS,
				Matrix_toEuler_doc},
	{"toQuat",(PyCFunction)Matrix_toQuat, METH_NOARGS,
				Matrix_toQuat_doc},	
	{NULL, NULL, 0, NULL}
};

//******************************************************************
//				Internal Matrix Functions
//******************************************************************
static PyObject *Matrix_toQuat(MatrixObject *self, PyObject *args)
{
	float *quat;

	quat = PyMem_Malloc(4*sizeof(float));
	Mat3ToQuat(*self->matrix,quat);

	return newQuaternionObject(quat);
}


static PyObject *Matrix_toEuler(MatrixObject *self, PyObject *args)
{
	float *eul;

	eul = PyMem_Malloc(3*sizeof(float));
	Mat3ToEul(*self->matrix,eul);

	return newEulerObject(eul);
}

static PyObject *Matrix_Resize4x4(MatrixObject *self, PyObject *args)
{
	float *mat;
	float * contigPtr;
	int x, row, col;

	mat = PyMem_Malloc(4*4*sizeof(float));
	for(x = 0; x < 15; x++){
		mat[x] = 0.0f;
	}
	mat[15] = 1.0f;

	if(self->colSize == 4 && self->rowSize == 4)
		return EXPP_incr_ret(Py_None);

	if(self->colSize == 2){ //2x2, 3x2, 4x2
			mat[0] = self->matrix[0][0];mat[1] = self->matrix[0][1];
			mat[4] = self->matrix[1][0];mat[5] = self->matrix[1][1];
		if (self->rowSize > 2){
			mat[8] = self->matrix[2][0];mat[9] = self->matrix[2][1];
		}
		if (self->rowSize > 3){
			mat[12] = self->matrix[3][0];mat[13] = self->matrix[3][1];
		}
	}else if (self->colSize == 3){  //2x3, 3x3, 4x3
			mat[0] = self->matrix[0][0];mat[1] = self->matrix[0][1];
			mat[2] = self->matrix[0][2];mat[4] = self->matrix[1][0];
			mat[5] = self->matrix[1][1];mat[6] = self->matrix[1][2];
		if (self->rowSize > 2){
			mat[8] = self->matrix[2][0];mat[9] = self->matrix[2][1];
			mat[10] = self->matrix[2][2];
		}
		if (self->rowSize > 3){
			mat[12] = self->matrix[3][0];mat[13] = self->matrix[3][1];
			mat[14] = self->matrix[3][2];
		}
	}else if (self->colSize == 4){  //2x4, 3x4
			mat[0] = self->matrix[0][0];mat[1] = self->matrix[0][1];
			mat[2] = self->matrix[0][2];mat[3] = self->matrix[0][3];
			mat[4] = self->matrix[1][0];mat[5] = self->matrix[1][1];
			mat[6] = self->matrix[1][2];mat[7] = self->matrix[1][3];
		if (self->rowSize > 2
			){
			mat[8] = self->matrix[2][0];mat[9] = self->matrix[2][1];
			mat[10] = self->matrix[2][2];mat[11] = self->matrix[2][3];
		}
	}

	PyMem_Free(*self->matrix);
	contigPtr = PyMem_Malloc(4 * 4 * sizeof(float));
	if(contigPtr == NULL){
		return (EXPP_ReturnPyObjError (PyExc_MemoryError,
				"problem allocating array space\n\n"));
	}
	self->matrix = PyMem_Malloc(4* sizeof(float*));
	if(self->matrix == NULL){
		return (EXPP_ReturnPyObjError (PyExc_MemoryError,
				"problem allocating pointer space\n\n"));
	}
    for (x = 0; x < 4; x++){
        self->matrix[x] = contigPtr + (x *4);
    }

	for (row = 0; row < 4; row++){
		for(col = 0; col < 4; col++){
			self->matrix[row][col] = mat[(row * 4) + col];
		}
	}
	PyMem_Free(mat);

	self->colSize = 4;
	self->rowSize = 4;

	return EXPP_incr_ret(Py_None);
}

static PyObject *Matrix_TranslationPart(MatrixObject *self, PyObject *args)
{
	float *vec;

	if(self->colSize < 3){
		return EXPP_ReturnPyObjError(PyExc_AttributeError,
		  "inappropriate matrix size\n");
	}else if (self->colSize > 2){    //3 or 4 columns
		if(self->rowSize < 4)		//all 4 rows
			return EXPP_ReturnPyObjError(PyExc_AttributeError,
			  "inappropriate matrix size\n");

		vec = PyMem_Malloc(3 * sizeof(float));
		vec[0] = self->matrix[3][0];
		vec[1] = self->matrix[3][1];
		vec[2] = self->matrix[3][2];
	}

	return newVectorObject(vec,3);
}

static PyObject *Matrix_RotationPart(MatrixObject *self, PyObject *args)
{
	float *mat;

	if(self->colSize < 3){
		return EXPP_ReturnPyObjError(PyExc_AttributeError,
		  "inappropriate matrix size\n");
	}else if (self->colSize > 2){	//3 or 4 col
		if(self->rowSize < 3)		//3 or 4 row
			return EXPP_ReturnPyObjError(PyExc_AttributeError,
			  "inappropriate matrix size\n");

		mat = PyMem_Malloc(3*3*sizeof(float));
		mat[0] = self->matrix[0][0];
		mat[1] = self->matrix[0][1];
		mat[2] = self->matrix[0][2];
		mat[3] = self->matrix[1][0];
		mat[4] = self->matrix[1][1];
		mat[5] = self->matrix[1][2];
		mat[6] = self->matrix[2][0];
		mat[7] = self->matrix[2][1];
		mat[8] = self->matrix[2][2];
	}

	return newMatrixObject(mat,3,3);
}

static PyObject *Matrix_Inverse(MatrixObject *self, PyObject *args)
{
	float det;
	int x,y,z;
	float *mat;
	float t;

	if(self->rowSize != self->colSize)
		return EXPP_ReturnPyObjError(PyExc_AttributeError,
		  "only square matrices are supported\n");

	//calculate the determinant
	if(self->rowSize == 2){
		det =  Det2x2(self->matrix[0][0],
			          self->matrix[0][1],
					  self->matrix[1][0],
					  self->matrix[1][1]);
	}else if(self->rowSize == 3){
		det =  Det3x3(self->matrix[0][0],
			          self->matrix[0][1],
					  self->matrix[0][2],
					  self->matrix[1][0],
					  self->matrix[1][1],
					  self->matrix[1][2],
					  self->matrix[2][0],
					  self->matrix[2][1],
					  self->matrix[2][2]);
	}else if(self->rowSize == 4){
		det =  Det4x4(*self->matrix);
	}else{
		return EXPP_ReturnPyObjError(PyExc_StandardError,
		  "error calculating determinant for inverse()\n");
	}

	if(det != 0){

		//calculate the classical adjoint
		if(self->rowSize == 2){
			mat = PyMem_Malloc(self->rowSize * self->colSize * sizeof(float));
			mat[0] = self->matrix[1][1];
			mat[1] = -self->matrix[1][0];
			mat[2] = -self->matrix[0][1];
			mat[3] = self->matrix[0][0];
		}else if(self->rowSize == 3){
			mat = PyMem_Malloc(self->rowSize * self->colSize * sizeof(float));
			Mat3Adj(mat, *self->matrix);
		}else if (self->rowSize == 4){
			mat = PyMem_Malloc(self->rowSize * self->colSize * sizeof(float));
			Mat4Adj(mat, *self->matrix);
		}

		//divide by determinate
		for(x = 0; x < (self->rowSize * self->colSize); x++){
			mat[x] /= det;
		}

		//set values
		z = 0.0f;
		for(x = 0; x < self->rowSize; x++){
			for(y = 0; y < self->colSize; y++){
				self->matrix[x][y] = mat[z];
				z++;
			}
		}

		//transpose
		if(self->rowSize == 2){
			t = self->matrix[1][0];
			self->matrix[1][0] = self->matrix[0][1];
			self->matrix[0][1] = t;
		}else if(self->rowSize == 3){
			Mat3Transp(mat);
		}
		else if (self->rowSize == 4){
			Mat4Transp(mat);
		}
	}else{
		printf("This matrix does not have an inverse - no inverse attempted");
	}

	return EXPP_incr_ret(Py_None);
}


static PyObject *Matrix_Determinant(MatrixObject *self, PyObject *args)
{
	float det;

	if(self->rowSize != self->colSize)
		return EXPP_ReturnPyObjError(PyExc_AttributeError,
		  "only square matrices are supported\n");

	if(self->rowSize == 2){
		det =  Det2x2(self->matrix[0][0],
			          self->matrix[0][1],
					  self->matrix[1][0],
					  self->matrix[1][1]);
	}else if(self->rowSize == 3){
		det =  Det3x3(self->matrix[0][0],
			          self->matrix[0][1],
					  self->matrix[0][2],
					  self->matrix[1][0],
					  self->matrix[1][1],
					  self->matrix[1][2],
					  self->matrix[2][0],
					  self->matrix[2][1],
					  self->matrix[2][2]);
	}else if(self->rowSize == 4){
		det =  Det4x4(*self->matrix);
	}else{
		return EXPP_ReturnPyObjError(PyExc_StandardError,
		  "error in determinant()\n");
	}

	return PyFloat_FromDouble((double)(det));
}

static PyObject *Matrix_Transpose(MatrixObject *self, PyObject *args)
{
	float t;

	if(self->rowSize != self->colSize)
		return EXPP_ReturnPyObjError(PyExc_AttributeError,
		  "only square matrices are supported\n");

	if(self->rowSize == 2){
		t = self->matrix[1][0];
		self->matrix[1][0] = self->matrix[0][1];
		self->matrix[0][1] = t;
	}
	else if(self->rowSize == 3){
		Mat3Transp(*self->matrix);
	}
	else if (self->rowSize == 4){
		Mat4Transp(*self->matrix);
	}
	else
		return (EXPP_ReturnPyObjError (PyExc_TypeError,
						"unable to transpose matrix\n"));

	return EXPP_incr_ret(Py_None);
}

static PyObject *Matrix_Zero(MatrixObject *self, PyObject *args)
{
	int row, col;

	for (row = 0; row < self->rowSize; row++){
		for (col = 0; col < self->colSize; col++){
			self->matrix[row][col] = 0.0f;
		}
	}
	return EXPP_incr_ret(Py_None);
}

static PyObject *Matrix_Identity(MatrixObject *self, PyObject *args)
{
	if(self->rowSize != self->colSize)
		return (EXPP_ReturnPyObjError(PyExc_AttributeError,
			"only square matrices supported\n"));

	if(self->rowSize == 2){
		self->matrix[0][0] = 1.0f;
		self->matrix[0][1] = 0.0f;
		self->matrix[1][0] = 0.0f;
		self->matrix[1][1] = 1.0f;
	}
	else if(self->rowSize == 3){
		Mat3One(*self->matrix);
	}
	else if (self->rowSize == 4){
		Mat4One(*self->matrix);
	}
	else
		return (EXPP_ReturnPyObjError (PyExc_TypeError,
						"unable to create identity matrix\n"));

	return EXPP_incr_ret(Py_None);
}

static void Matrix_dealloc (MatrixObject *self)
{
	PyMem_Free (self->matrix);
    PyMem_DEL (self);
}

static PyObject * Matrix_getattr (MatrixObject *self, char *name)
{
    if (strcmp (name, "rowSize") == 0){
		return PyInt_FromLong((long)self->rowSize);
    }
    else if (strcmp (name, "colSize") == 0){
		return PyInt_FromLong((long)self->colSize);
    }

	return Py_FindMethod(Matrix_methods, (PyObject*)self, name);
}

static int Matrix_setattr (MatrixObject *self, char *name, PyObject *v)
{
    /* This is not supported. */
    return (-1);
}

static PyObject * Matrix_repr (MatrixObject *self)
{
	PyObject *repr, *str;
	int x,y;
	char ftoa[24];

	repr = PyString_FromString("");
	if (!repr) 
		return (EXPP_ReturnPyObjError (PyExc_AttributeError,
                "Attribute error in PyMatrix (repr)\n"));

	for(x = 0; x < self->rowSize; x++){
		str = PyString_FromString ("[");
		PyString_ConcatAndDel(&repr,str);

		for(y = 0; y < (self->colSize - 1); y++){
			sprintf(ftoa, "%.4f, ", self->matrix[x][y]);
			str = PyString_FromString (ftoa);
			PyString_ConcatAndDel(&repr,str);
		}
		sprintf(ftoa, "%.4f]\n", self->matrix[x][y]);
		str = PyString_FromString (ftoa);
		PyString_ConcatAndDel(&repr,str);
	}
	return repr;
}

//no support for matrix[x][y] so have to return by sequence index
static PyObject * Matrix_item (MatrixObject *self, int i)
{
	int maxsize, x, y;

	maxsize = self->colSize * self->rowSize;
	if(i < 0 || i >= maxsize)
		return EXPP_ReturnPyObjError(PyExc_IndexError,
			"array index out of range\n");

	x = floor((double)(i / self->colSize));
	y = i % self->colSize;
 
	return PyFloat_FromDouble((double)self->matrix[x][y]);
}

PyObject * Matrix_add(PyObject *v, PyObject * w)
{
	float * vec;
	int matSize, rowSize, colSize, x,y;

	if((!Matrix_CheckPyObject(v)) || (!Matrix_CheckPyObject(w)))
		return EXPP_ReturnPyObjError (PyExc_TypeError,
				"unsupported type for this operation\n");

	if(((MatrixObject*)w)->flag != 0)
		return EXPP_ReturnPyObjError (PyExc_ArithmeticError,
			"cannot add scalar to a matrix\n");

	rowSize = (((MatrixObject*)v)->rowSize);
	colSize = (((MatrixObject*)v)->colSize);
	matSize = rowSize * colSize;

	if(rowSize != ((MatrixObject*)w)->rowSize &&
		colSize != ((MatrixObject*)w)->colSize)
		return EXPP_ReturnPyObjError (PyExc_AttributeError,
				"Matrices must be the same same for this operation\n");
	
	vec = PyMem_Malloc (matSize * sizeof(float));

	for(x = 0; x < rowSize; x++){
		for(y = 0; y < colSize; y++){
			vec[((x * rowSize) + y)] = 
				((MatrixObject*)v)->matrix[x][y] +
				((MatrixObject*)w)->matrix[x][y];
		}
	}

	return newMatrixObject(vec, rowSize, colSize);
}

PyObject * Matrix_sub(PyObject *v, PyObject * w)
{
	float * vec;
	int matSize, rowSize, colSize, x,y;

	if((!Matrix_CheckPyObject(v)) || (!Matrix_CheckPyObject(w)))
		return EXPP_ReturnPyObjError (PyExc_TypeError,
				"unsupported type for this operation\n");

	if(((MatrixObject*)w)->flag != 0)
		return EXPP_ReturnPyObjError (PyExc_ArithmeticError,
			"cannot add scalar to a matrix\n");

	rowSize = (((MatrixObject*)v)->rowSize);
	colSize = (((MatrixObject*)v)->colSize);
	matSize = rowSize * colSize;

	if(rowSize != ((MatrixObject*)w)->rowSize &&
		colSize != ((MatrixObject*)w)->colSize)
		return EXPP_ReturnPyObjError (PyExc_AttributeError,
				"Matrices must be the same same for this operation\n");
	
	vec = PyMem_Malloc (matSize * sizeof(float));

	for(x = 0; x < rowSize; x++){
		for(y = 0; y < colSize; y++){
			vec[((x * rowSize) + y)] = 
				((MatrixObject*)v)->matrix[x][y] -
				((MatrixObject*)w)->matrix[x][y];
		}
	}

	return newMatrixObject(vec, rowSize, colSize);
}

PyObject * Matrix_mul(PyObject *v, PyObject * w)
{
	float * vec;
	int matSizeV, rowSizeV, colSizeV, rowSizeW, colSizeW, matSizeW, x, y, z;
	float dot = 0;
	MatrixObject * matV;
	MatrixObject * matW;

	if((!Matrix_CheckPyObject(v)) || (!Matrix_CheckPyObject(w)))
		return EXPP_ReturnPyObjError (PyExc_TypeError,
				"unsupported type for this operation\n");

	//get some vars
	rowSizeV = (((MatrixObject*)v)->rowSize);
	colSizeV = (((MatrixObject*)v)->colSize);
	matSizeV = rowSizeV * colSizeV;
	rowSizeW = (((MatrixObject*)w)->rowSize);
	colSizeW = (((MatrixObject*)w)->colSize);
	matSizeW = rowSizeW * colSizeW;
	matV = ((MatrixObject*)v);
	matW = ((MatrixObject*)w);

	//coerced int or float for scalar multiplication
	if(matW->flag == 1 || matW->flag == 2){

		if(rowSizeV != rowSizeW &&	colSizeV != colSizeW)
			return EXPP_ReturnPyObjError (PyExc_AttributeError,
					"Matrix dimension error during scalar multiplication\n");
		
		vec = PyMem_Malloc (matSizeV * sizeof(float));

		for(x = 0; x < rowSizeV; x++){
			for(y = 0; y < colSizeV; y++){
				vec[((x * rowSizeV) + y)] = 
					matV->matrix[x][y] * matW->matrix[x][y];
			}
		}
		return newMatrixObject(vec, rowSizeV, colSizeV);
	}
	else if (matW->flag == 0){		//true matrix multiplication
		if(colSizeV != rowSizeW){
			return EXPP_ReturnPyObjError (PyExc_AttributeError,
					"Matrix multiplication undefined...\n");
		}

		vec = PyMem_Malloc((rowSizeV * colSizeW) * sizeof(float));

		for(x = 0; x < rowSizeV; x++){
			for(y = 0; y < colSizeW; y++){
				for(z = 0; z < colSizeV; z++){
					dot += (matV->matrix[x][z] * matW->matrix[z][y]);
				}
				vec[((x * rowSizeV) + y)] = dot;
				dot = 0;
			}
		}
		return newMatrixObject(vec, rowSizeV, colSizeW);
	}
	else
		return EXPP_ReturnPyObjError (PyExc_AttributeError,
					"Error in matrix_mul...\n");
}

//coercion of unknown types to type MatrixObject for numeric protocols
int Matrix_coerce(PyObject **v, PyObject **w)
{ 
	long *tempI;
	double *tempF;
	float *vec;
	int x, matSize, y;

	matSize = (((MatrixObject*)*v)->rowSize) * (((MatrixObject*)*v)->rowSize);
	if (Matrix_CheckPyObject(*v)) {
		//1st is Matrix
		if (Matrix_CheckPyObject(*w)) {
			//2nd is Matrix`- this shouldn't get called - but just in case....
			Py_INCREF(*v);
			Py_INCREF(*w);
			return 0;
		}else{
			//2nd not Matrix
			if(PyNumber_Check(*w)){
				//it's a number
				if(PyInt_Check(*w)){
					//it's a int
					tempI = PyMem_Malloc(1*sizeof(long));
					*tempI = PyInt_AsLong(*w);
					vec = PyMem_Malloc (matSize * sizeof (float));
					for(x = 0; x < matSize; x++){
						vec[x] = *tempI;
					}
					PyMem_Free(tempI);
					*w = newMatrixObject(vec, (((MatrixObject*)*v)->rowSize),
						(((MatrixObject*)*v)->colSize));
					((MatrixObject*)*w)->flag = 1;	//int coercion
					PyMem_Free(vec);
					Py_INCREF(*v);
					return 0;
				}else if(PyFloat_Check(*w)){
					//it's a float
					tempF = PyMem_Malloc(1*sizeof(double));
					*tempF = PyFloat_AsDouble(*w);
					vec = PyMem_Malloc (matSize * sizeof (float));
					for(x = 0; x < matSize; x++){
						vec[x] = *tempF;
					}
					PyMem_Free(tempF);
					*w = newMatrixObject(vec, (((MatrixObject*)*v)->rowSize),
						(((MatrixObject*)*v)->colSize));
					((MatrixObject*)*w)->flag = 2;	//float coercion
					PyMem_Free(vec);
					Py_INCREF(*v);
					return 0;
				}else{
					//not a number - maybe a vector? 
					if(VectorObject_Check(*w)){
						printf("please use the MatMultVec() function...\n");
						return -1;
					}else{
						printf("attempting operation with unsupported type...\n");
						return -1;
					}
				}
			}else{
				//not a number - fail
				printf("attempting operation with non numeric type...\n");
				return -1;
			}
		}
	}else{
		//1st not Matrix
		printf("coercion failure...\n");
		return -1; //this should not occur - fail
	}
	return -1; 
}

//******************************************************************
//					Matrix definition
//******************************************************************
static PySequenceMethods Matrix_SeqMethods =
{
    (inquiry) 0,                /*sq_length*/
    (binaryfunc) 0,             /*sq_concat*/
    (intargfunc) 0,             /*sq_repeat*/
    (intargfunc) Matrix_item,   /*sq_item*/
    (intintargfunc) 0,          /*sq_slice*/
    (intobjargproc) 0,          /*sq_ass_item*/
    (intintobjargproc) 0,       /*sq_ass_slice*/
};

static PyNumberMethods Matrix_NumMethods =
{
    (binaryfunc)	Matrix_add,               /* __add__ */
    (binaryfunc)	Matrix_sub,               /* __sub__ */
    (binaryfunc)	Matrix_mul,               /* __mul__ */
    (binaryfunc)	0,       		          /* __div__ */
    (binaryfunc)	0,				          /* __mod__ */
    (binaryfunc)	0,                        /* __divmod__ */
    (ternaryfunc)	0,                        /* __pow__ */
    (unaryfunc)		0,                        /* __neg__ */
    (unaryfunc)		0,                        /* __pos__ */
    (unaryfunc)		0,                        /* __abs__ */
    (inquiry)		0,                        /* __nonzero__ */
    (unaryfunc)		0,                        /* __invert__ */
    (binaryfunc)	0,                        /* __lshift__ */
    (binaryfunc)	0,                        /* __rshift__ */
    (binaryfunc)	0,                        /* __and__ */
    (binaryfunc)	0,                        /* __xor__ */
    (binaryfunc)	0,                        /* __or__ */
    (coercion)		Matrix_coerce,			  /* __coerce__ */
    (unaryfunc)		0,                        /* __int__ */
    (unaryfunc)		0,                        /* __long__ */
    (unaryfunc)		0,                        /* __float__ */
    (unaryfunc)		0,                        /* __oct__ */
    (unaryfunc)		0,                        /* __hex__ */
};

PyTypeObject matrix_Type =
{
    PyObject_HEAD_INIT(NULL)
    0,                              /*ob_size*/
    "Matrix",                       /*tp_name*/
    sizeof(MatrixObject),           /*tp_basicsize*/
    0,                              /*tp_itemsize*/
    /* methods */
    (destructor)    Matrix_dealloc, /*tp_dealloc*/
    (printfunc)     0,              /*tp_print*/
    (getattrfunc)   Matrix_getattr, /*tp_getattr*/
    (setattrfunc)   Matrix_setattr, /*tp_setattr*/
    0,                              /*tp_compare*/
    (reprfunc)      Matrix_repr,    /*tp_repr*/
    &Matrix_NumMethods,             /*tp_as_number*/
    &Matrix_SeqMethods,             /*tp_as_sequence*/
};

void init_py_matrix (void)
{
    matrix_Type.ob_type = &PyType_Type;
}

//******************************************************************
//Function:				 newMatrixObject
//******************************************************************
PyObject * newMatrixObject (float * mat, int rowSize, int colSize)
{
    MatrixObject    * self;
	float * contigPtr;
	int row, col, x;

	if (rowSize < 2 || rowSize > 4 || colSize < 2 || colSize > 4)
		return (EXPP_ReturnPyObjError (PyExc_RuntimeError,
                "row and column sizes must be between 2 and 4\n"));

    self = PyObject_NEW (MatrixObject, &matrix_Type);

	//generate contigous memory space
	contigPtr = PyMem_Malloc(rowSize * colSize* sizeof(float));
	if(contigPtr == NULL){
		return (EXPP_ReturnPyObjError (PyExc_MemoryError,
				"problem allocating array space\n\n"));
	}

	//create pointer array
	self->matrix = PyMem_Malloc(rowSize * sizeof(float*));
	if(self->matrix == NULL){
		return (EXPP_ReturnPyObjError (PyExc_MemoryError,
				"problem allocating pointer space\n\n"));
	}

	//pointer array points to contigous memory
    for (x = 0; x < rowSize; x++){
        self->matrix[x] = contigPtr + (x * colSize);
    }

	if(mat){	//if a float array passed
		for (row = 0; row < rowSize; row++){
			for(col = 0; col < colSize; col++){
				self->matrix[row][col] = mat[(row * colSize) + col];
			}
		}
	}else{		//or if NULL passed
		for (row = 0; row < rowSize; row++){
			for (col = 0; col < colSize; col++){
				self->matrix[row][col] = 0.0f;
			}
		}
	}

	//set size vars of matrix
	self->rowSize = rowSize;
	self->colSize = colSize;

	//set coercion flag
	self->flag = 0;

    return ((PyObject *)self);
}

//***************************************************************************
// Function:		Matrix_CheckPyObject prototypes																 
//***************************************************************************
int Matrix_CheckPyObject (PyObject *pyobj)
{
	return (pyobj->ob_type == &matrix_Type);
}

