/*
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * 
 * Contributor(s): Joseph Gilbert
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

#include "quat.h"
#include "Types.h"

/*****************************/
//    Quaternion Python Object   
/*****************************/

struct PyMethodDef Quaternion_methods[] = {
	{"identity",(PyCFunction)Quaternion_Identity, METH_NOARGS,
				Quaternion_Identity_doc},
	{"negate",(PyCFunction)Quaternion_Negate, METH_NOARGS,
				Quaternion_Negate_doc},
	{"conjugate",(PyCFunction)Quaternion_Conjugate, METH_NOARGS,
				Quaternion_Conjugate_doc},
	{"inverse",(PyCFunction)Quaternion_Inverse, METH_NOARGS,
				Quaternion_Inverse_doc},
	{"normalize",(PyCFunction)Quaternion_Normalize, METH_NOARGS,
				Quaternion_Normalize_doc},
	{"toEuler",(PyCFunction)Quaternion_ToEuler, METH_NOARGS,
				Quaternion_ToEuler_doc},
	{"toMatrix",(PyCFunction)Quaternion_ToMatrix, METH_NOARGS,
				Quaternion_ToMatrix_doc},
	{NULL, NULL, 0, NULL}
};

static PyObject *Quaternion_ToEuler(QuaternionObject *self, PyObject *args)
{
	float *eul;

	eul = PyMem_Malloc(3*sizeof(float));
	QuatToEul(self->vec, eul);

	return newEulerObject(eul);
}

static PyObject *Quaternion_ToMatrix(QuaternionObject *self, PyObject *args)
{
	float *mat;

	mat = PyMem_Malloc(3*3*sizeof(float));
	QuatToMat3(self->vec, mat);

	return newMatrixObject(mat, 3,3);
}

//normalize the axis of rotation of [theta,vector]
static PyObject *Quaternion_Normalize(QuaternionObject *self, PyObject *args)
{
	NormalQuat(self->vec);

	return EXPP_incr_ret(Py_None);
}

static PyObject *Quaternion_Inverse(QuaternionObject *self, PyObject *args)
{
	float mag;
	int x;

	self->vec[1] = -self->vec[1];
	self->vec[2] = -self->vec[2];
	self->vec[3] = -self->vec[3];

	mag = (float)sqrt((double)self->vec[0] * (double)self->vec[0] +
				(double)self->vec[1] * (double)self->vec[1] +
				(double)self->vec[2] * (double)self->vec[2] +
				(double)self->vec[3] * (double)self->vec[3]);

	for(x = 0; x < 4; x++){
		self->vec[x] /= (mag * mag);
	}

	return EXPP_incr_ret(Py_None);
}

static PyObject *Quaternion_Identity(QuaternionObject *self, PyObject *args)
{
	self->vec[0] = 1.0;
	self->vec[1] = 0.0; self->vec[2] = 0.0; self->vec[3] = 0.0;

	return EXPP_incr_ret(Py_None);
}

static PyObject *Quaternion_Negate(QuaternionObject *self, PyObject *args)
{
	int x;

	for(x = 0; x < 4; x++){
		self->vec[x] = -self->vec[x];
	}

	return EXPP_incr_ret(Py_None);
}

static PyObject *Quaternion_Conjugate(QuaternionObject *self, PyObject *args)
{
	int x;

	for(x = 1; x < 4; x++){
		self->vec[x] = -self->vec[x];
	}

	return EXPP_incr_ret(Py_None);
}

static void Quaternion_dealloc(QuaternionObject *self)
{
  PyObject_DEL (self);
}

static PyObject *Quaternion_getattr(QuaternionObject *self, char *name)
{
	double mag;
	float *vec;
	int x;

	if (ELEM4(name[0], 'w', 'x', 'y', 'z') && name[1]==0){
		return PyFloat_FromDouble(self->vec[name[0]-'x']);
	}

	if(strcmp(name,"magnitude") == 0){
		mag = sqrt((double)self->vec[0] * (double)self->vec[0] +
				   (double)self->vec[1] * (double)self->vec[1] +
				   (double)self->vec[2] * (double)self->vec[2] +
				   (double)self->vec[3] * (double)self->vec[3]);
		return PyFloat_FromDouble(mag);
	}
	if(strcmp(name,"angle") == 0){

		mag = (double)(self->vec[0] * (M_PI/180));
		mag = 2 * (acos(mag));

		return PyFloat_FromDouble(mag);
	}
	if(strcmp(name,"axis") == 0){

		mag = (double)(self->vec[0] * (M_PI/180));
		mag = 2 * (acos(mag));
		mag = sin(mag/2);

		vec = PyMem_Malloc(3*sizeof(float));

		for(x = 0; x < 3; x++){
			vec[x] = (self->vec[x + 1]/((float)(mag)));
		}

		return newVectorObject(vec,3);
	}

	return Py_FindMethod(Quaternion_methods, (PyObject*)self, name);
}

static int Quaternion_setattr(QuaternionObject *self, char *name, PyObject *v)
{
	float val;
	int valTemp;

	if(!PyFloat_Check(v)){
		if(!PyInt_Check(v)){
			return EXPP_ReturnIntError(PyExc_TypeError,"int or float expected\n");
		}else{
			if (!PyArg_Parse(v, "i", &valTemp))
				return EXPP_ReturnIntError(PyExc_TypeError,	"unable to parse int argument\n");
			val = (float)valTemp;
		}
	}else{
		if (!PyArg_Parse(v, "f", &val))
			return EXPP_ReturnIntError(PyExc_TypeError,	"unable to parse float argument\n");
	}
	if (ELEM4(name[0], 'w', 'x', 'y', 'z') && name[1]==0){
		self->vec[name[0]-'x']= val;
	}
	else return -1;

	return 0;
}

/* Quaternions Sequence methods */
static PyObject *Quaternion_item(QuaternionObject *self, int i)
{
	if (i < 0 || i >= 4)
	  return EXPP_ReturnPyObjError (PyExc_IndexError, "array index out of range\n");

	return Py_BuildValue("f", self->vec[i]);

}

static PyObject *Quaternion_slice(QuaternionObject *self, int begin, int end)
{
	PyObject *list;
	int count;
  
	if (begin < 0) begin= 0;
	if (end > 4) end= 4;
	if (begin > end) begin= end;

	list= PyList_New(end-begin);

	for (count = begin; count < end; count++){
		PyList_SetItem(list, count-begin, PyFloat_FromDouble(self->vec[count]));
	}

	return list;
}

static int Quaternion_ass_item(QuaternionObject *self, int i, PyObject *ob)
{
	int temp;

	if (i < 0 || i >= 4)
		return EXPP_ReturnIntError(PyExc_IndexError,
					"array assignment index out of range\n");

	if (!PyNumber_Check(ob))
		return EXPP_ReturnIntError(PyExc_IndexError,
					"Quaternion member must be a number\n");

	if(!PyFloat_Check(ob)){
		if(!PyInt_Check(ob)){
			return EXPP_ReturnIntError(PyExc_TypeError,"int or float expected\n");
		}else{
			temp = PyInt_AsLong(ob);
			self->vec[i] = ((float)temp);
		}
	}else{
		self->vec[i]= PyFloat_AsDouble(ob);
	}

	return 0;
}

static int Quaternion_ass_slice(QuaternionObject *self, int begin, int end, PyObject *seq)
{
	int count;
	int tempInt;

	if (begin < 0) begin= 0;
	if (end > 4) end= 4;
	if (begin > end) begin= end;

	if (!PySequence_Check(seq))
		return EXPP_ReturnIntError(PyExc_TypeError,
					"illegal argument type for built-in operation\n");

	if (PySequence_Length(seq) != (end - begin))
		return EXPP_ReturnIntError(PyExc_TypeError,
					"size mismatch in slice assignment\n");

	for (count = begin; count < end; count++) {
		PyObject *ob = PySequence_GetItem(seq, count);

		if(!EXPP_check_sequence_consistency(ob, &PyFloat_Type)){
			if(!EXPP_check_sequence_consistency(ob, &PyInt_Type)){
				return -1;
			}else{
				if (!PyArg_Parse(ob, "i", &tempInt)) {
					self->vec[count] = ((float)tempInt);
					Py_DECREF(ob);
					return -1;
				}
			}
		}else{
			if (!PyArg_Parse(ob, "f", &self->vec[count])) {
				Py_DECREF(ob);
				return -1;
			}
		}

		Py_DECREF(ob);
	}

  return 0;
}

static PyObject *Quaternion_repr (QuaternionObject *self)
{
	int i, maxindex = 4 - 1;
	char ftoa[24];
	PyObject *str1, *str2;
	int x;

	str1 = PyString_FromString ("[");

	for (i = 0; i < maxindex; i++) {
		sprintf(ftoa, "%.3f, ", self->vec[i]);
		str2 = PyString_FromString (ftoa);
		if (!str1 || !str2) goto error; 
		PyString_ConcatAndDel (&str1, str2);
	}

	sprintf(ftoa, "%.3f]\n", self->vec[maxindex]);
	str2 = PyString_FromString (ftoa);
	if (!str1 || !str2) goto error; 
	PyString_ConcatAndDel (&str1, str2);

	if (str1) return str1;

error:
	Py_XDECREF (str1);
	Py_XDECREF (str2);
	return EXPP_ReturnPyObjError (PyExc_MemoryError,
			"couldn't create PyString!\n");
}


PyObject * Quaternion_add(PyObject *v, PyObject *w)
{
	float * quat;

	if((!QuaternionObject_Check(v)) || (!QuaternionObject_Check(w)))
		return EXPP_ReturnPyObjError (PyExc_TypeError,
				"unsupported type for this operation\n");

	if(((QuaternionObject*)w)->flag != 0)
		return EXPP_ReturnPyObjError (PyExc_ArithmeticError,
			"cannot add a scalar and a quat\n");

	quat = PyMem_Malloc (4*sizeof(float));
	quat[0] =  (((QuaternionObject*)v)->vec[0]) +  (((QuaternionObject*)w)->vec[0]);
	quat[1] =  (((QuaternionObject*)v)->vec[1]) +  (((QuaternionObject*)w)->vec[1]);
	quat[2] =  (((QuaternionObject*)v)->vec[2]) +  (((QuaternionObject*)w)->vec[2]);
	quat[3] =  (((QuaternionObject*)v)->vec[3]) +  (((QuaternionObject*)w)->vec[3]);

	return (PyObject*)newQuaternionObject(quat);
}

PyObject * Quaternion_sub(PyObject *v, PyObject *w)
{
	float * quat;

	if((!QuaternionObject_Check(v)) || (!QuaternionObject_Check(w)))
		return EXPP_ReturnPyObjError (PyExc_TypeError,
				"unsupported type for this operation\n");

	if(((QuaternionObject*)w)->flag != 0)
		return EXPP_ReturnPyObjError (PyExc_ArithmeticError,
			"cannot subtract a scalar and a quat\n");

	quat = PyMem_Malloc (4*sizeof(float));
	quat[0] =  (((QuaternionObject*)v)->vec[0]) -  (((QuaternionObject*)w)->vec[0]);
	quat[1] =  (((QuaternionObject*)v)->vec[1]) -  (((QuaternionObject*)w)->vec[1]);
	quat[2] =  (((QuaternionObject*)v)->vec[2]) -  (((QuaternionObject*)w)->vec[2]);
	quat[3] =  (((QuaternionObject*)v)->vec[3]) -  (((QuaternionObject*)w)->vec[3]);

	return (PyObject*)newQuaternionObject(quat);
}

PyObject * Quaternion_mul(PyObject *v, PyObject * w)
{
	float * quat;
	int x;

	if((!QuaternionObject_Check(v)) || (!QuaternionObject_Check(w)))
		return EXPP_ReturnPyObjError (PyExc_TypeError,
				"unsupported type for this operation\n");

	if(((QuaternionObject*)w)->flag == 0)
		return EXPP_ReturnPyObjError (PyExc_ArithmeticError,
			"please use the dot product or the cross product to multiply quaternions\n");

	if(((QuaternionObject*)w)->flag != 1 && ((QuaternionObject*)w)->flag != 2)
				return EXPP_ReturnPyObjError (PyExc_TypeError,
					"only scalar multiplication supported currently\n");

	quat = PyMem_Malloc (4*sizeof(float));

	//scalar mult by quat
	for(x = 0; x < 4; x++){
		quat[x] = ((QuaternionObject*)v)->vec[x] * ((QuaternionObject*)w)->vec[x];
	}

	return (PyObject*)newQuaternionObject(quat);
}

//quat exponentiation - useful for interpolation of quats
PyObject * Quaternion_pow(PyObject *o1, PyObject *o2, PyObject *o3)
{
	float angle, delta, theta;
	float * quat;

	if((!QuaternionObject_Check(o1)) || (!QuaternionObject_Check(o2)))
		return EXPP_ReturnPyObjError (PyExc_TypeError,
				"unsupported type for this operation\n");

	if(((QuaternionObject*)o2)->flag == 0)
		return EXPP_ReturnPyObjError (PyExc_ArithmeticError,
			"second argument must be a number\n");

	quat = PyMem_Malloc(4*sizeof(float));

	angle = (float)(acos((((QuaternionObject*)o1)->vec[0])));
	delta = angle * (((QuaternionObject*)o2)->vec[0]);
	quat[0] = (float)(cos(delta));
	theta = (float)(sin(delta)/sin(angle));
	quat[1] = ((QuaternionObject*)o1)->vec[1] * theta;
	quat[2] = ((QuaternionObject*)o1)->vec[2] * theta;
	quat[3] = ((QuaternionObject*)o1)->vec[3] * theta;

	return (PyObject*)newQuaternionObject(quat);
}

//coercion of unknown types to type QuaternionObject for numeric protocols
int Quaternion_coerce(PyObject **v, PyObject **w)
{ 
	long *tempI;
	double *tempF;
	float *vec;
	int x;

	if (QuaternionObject_Check(*v)) {
		//1st is Quaternion
		if (QuaternionObject_Check(*w)) {
			//2nd is Quaternion`- this shouldn't get called - but just in case....
			Py_INCREF(*v);
			Py_INCREF(*w);
			return 0;
		}else{
			//2nd not Quaternion
			if(PyNumber_Check(*w)){
				//it's a number
				if(PyInt_Check(*w)){
					//it's a int
					tempI = PyMem_Malloc(1*sizeof(long));
					*tempI = PyInt_AsLong(*w);
					vec = PyMem_Malloc (4*sizeof (float));
					for(x = 0; x < 4; x++){
						vec[x] = *tempI;
					}
					PyMem_Free(tempI);
					*w = newQuaternionObject(vec);
					((QuaternionObject*)*w)->flag = 1;	//int coercion
					Py_INCREF(*v);
					return 0;
				}else if(PyFloat_Check(*w)){
					//it's a float
					tempF = PyMem_Malloc(1*sizeof(double));
					*tempF = PyFloat_AsDouble(*w);
					vec = PyMem_Malloc (4*sizeof (float));
					for(x = 0; x < 4; x++){
						vec[x] = *tempF;
					}
					PyMem_Free(tempF);
					*w = newQuaternionObject(vec);
					((QuaternionObject*)*w)->flag = 2;	//float coercion
					Py_INCREF(*v);
					return 0;
				}else{
					//non supported numeric type
					printf("attempting operation with unsupported numeric type...\n");
					return -1;
				}
			}else{
     				//unknown
					printf("attempting operation with unknown type...\n");
					return -1;
			}
		}
	}else{
		//1st not Quaternion
		printf("coercion failure...\n");
		return -1; //this should not occur - fail
	}
	return -1; 
}


static PySequenceMethods Quaternion_SeqMethods =
{
	(inquiry)       0,                          /* sq_length */
	(binaryfunc)    0,                          /* sq_concat */
	(intargfunc)    0,                          /* sq_repeat */
	(intargfunc)    Quaternion_item,            /* sq_item */
	(intintargfunc)   Quaternion_slice,         /* sq_slice */
	(intobjargproc)   Quaternion_ass_item,      /* sq_ass_item */
	(intintobjargproc)  Quaternion_ass_slice,   /* sq_ass_slice */
};

static PyNumberMethods Quaternion_NumMethods =
{
    (binaryfunc)	Quaternion_add,           /* __add__ */
    (binaryfunc)	Quaternion_sub,           /* __sub__ */
    (binaryfunc)	Quaternion_mul,           /* __mul__ */
    (binaryfunc)	0,		                  /* __div__ */
    (binaryfunc)	0,				          /* __mod__ */
    (binaryfunc)	0,                        /* __divmod__ */
    (ternaryfunc)	Quaternion_pow,           /* __pow__ */
    (unaryfunc)		0,                        /* __neg__ */
    (unaryfunc)		0,                        /* __pos__ */
    (unaryfunc)		0,                        /* __abs__ */
    (inquiry)		0,                        /* __nonzero__ */
    (unaryfunc)		0,                        /* __invert__ */
    (binaryfunc)	0,                        /* __lshift__ */
    (binaryfunc)	0,                        /* __rshift__ */
    (binaryfunc)	0,                        /* __and__ */
    (binaryfunc)	0,                        /* __xor__ */
    (binaryfunc)	0,                        /* __or__ */
    (coercion)		Quaternion_coerce,        /* __coerce__ */
    (unaryfunc)		0,                        /* __int__ */
    (unaryfunc)		0,                        /* __long__ */
    (unaryfunc)		0,                        /* __float__ */
    (unaryfunc)		0,                        /* __oct__ */
    (unaryfunc)		0,                        /* __hex__ */

};

PyTypeObject quaternion_Type =
{
  PyObject_HEAD_INIT(NULL)
  0,                                /*ob_size*/
  "Quaternion",                    /*tp_name*/
  sizeof(QuaternionObject),        /*tp_basicsize*/
  0,                               /*tp_itemsize*/
  /* methods */
  (destructor)  Quaternion_dealloc, /*tp_dealloc*/
  (printfunc)   0,                  /*tp_print*/
  (getattrfunc) Quaternion_getattr, /*tp_getattr*/
  (setattrfunc) Quaternion_setattr, /*tp_setattr*/
  0,                                /*tp_compare*/
  (reprfunc)    Quaternion_repr,    /*tp_repr*/
  &Quaternion_NumMethods,           /*tp_as_number*/
  &Quaternion_SeqMethods,           /*tp_as_sequence*/
};

PyObject *newQuaternionObject(float *vec)
{
  QuaternionObject *self;
  int x;

  quaternion_Type.ob_type = &PyType_Type;

  self = PyObject_NEW(QuaternionObject, &quaternion_Type);

  if(!vec){
	  self->vec = PyMem_Malloc (4 *sizeof (float));
	  for(x = 0; x < 4; x++){
		  self->vec[x] = 0.0f;
	  }
	  self->vec[3] = 1.0f;
  }else{
	self->vec = vec;
  }
  self->flag = 0;
 
  return (PyObject*) self;
}

